/*  HSYFRINGES - Remove colour fringes using HSY colour space
 *
 * This file is part of PhotoPNMTools,
 * Release $Name: rel-1-3-0 $,
 * File rev. $Id: hsyfringes.c,v 1.1 2003/02/22 12:21:18 schooten Exp $.
 *
 * Copyright (c) 2002, 2003 by B.W. van Schooten. All rights reserved.
 * This software is distributed under the Perl Artistic License.
 * No warranty. See file 'artistic.license' for more details.
 *
 * boris@13thmonkey.org
 * www.13thmonkey.org/~boris/photopnmtools/
 */

#include <pnm.h>
#include <stdio.h>
#include <math.h>

#include "stdpnm.h"

#define USAGE \
"hsyfringes: remove blue fringes from digital camera pictures.\n" \
"The program requires input in PPM HSY format (i.e. ppmhsy -O).\n" \
"     -y [Y values above this are highlights, 0.0..1.0, default 0.5]\n" \
"     -h [hue of fringes, 0..1, default 0.6666]\n" \
"     -v [hue variance of fringes, 0..1, default 0.15]\n" \
"     -p [highlight pct. above which fringes occur, 0.0..1.0, default 0.25]\n" \
"     -r [blue reduction level, default 1.0]\n" \
"     -a [measuring area radius, > 1, default 5]\n" \
"\n" \
"     -X [xpos in pct. of bounding box, default 0]\n" \
"     -Y [ypos in pct. of bounding box, default 0]\n" \
"     -W [width pct. of bounding box, default 1]\n" \
"     -H [height pct. of bounding box, default 1]\n" \
"\n" \

#define ABS(a)  ((a) < 0 ? -(a) : (a) )


double ythresh = 0.5, highlightpct = 0.25, bluereduction=1;
double fringehue=0.6666, fringevar=0.15;
int area = 5;

double bx=0,by=0, bwidth=1,bheight=1;

double maxbright = 2.0;


	int
main(int argc, char **argv) {
	xel **orig,**final;
	struct rgb **in,**out;

	char optch;

	pixval maxpix;
	int format;
	int xsize,ysize;
	int x,y, xo,yo;

	int x1,y1,x2,y2;
	int hltpix,totpix;
	double hltlevel,hueaccuracy;
	pnm_init(&argc,argv);
	while ( (optch = getopt(argc, argv, "y:h:p:v:r:a:X:Y:W:H:")) != -1) {
		switch (optch) {
		case 'y': ythresh=atof(optarg);
		break;
		case 'h': fringehue=atof(optarg);
		break;
		case 'p': highlightpct=atof(optarg);
		break;
		case 'v': fringevar=atof(optarg);
		break;
		case 'r': bluereduction=atof(optarg);
		break;
		case 'a': area=atoi(optarg);
		break;
		case 'X': bx=atof(optarg);
		break;
		case 'Y': by=atof(optarg);
		break;
		case 'W': bwidth=atof(optarg);
		break;
		case 'H': bheight=atof(optarg);
		break;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc != 0) {
		fprintf(stderr,USAGE);
		exit(EXIT_FAILURE);
	}

	orig = pnm_readpnm(stdin,&xsize,&ysize,&maxpix,&format);
	in = rgballocarray(xsize, ysize, orig, maxpix);
	out = rgballocarray(xsize, ysize, NULL, maxpix);
	final = pnm_allocarray(xsize,ysize);

	x1 = bx * xsize;
	y1 = by * ysize;
	x2 = (bx+bwidth) * xsize;
	y2 = (by+bheight) * ysize;

	for (y=0; y<ysize; y++) {
		for (x=0; x<xsize; x++) {
			out[y][x].r = in[y][x].r;
			out[y][x].g = in[y][x].g;
			out[y][x].b = in[y][x].b;
			if (x >= x1 && y >= y1 && x <= x2 && y <= y2) {
				hltpix=0;
				totpix=0;
				for (yo=-area; yo<=area; yo++) {
					for (xo=-area; xo<=area; xo++) {
						if (x + xo >= 0    && y + yo >= 0
						&&  x + xo < xsize && y + yo < ysize) {
							if ((in[y+yo][x+xo].b*maxbright) > ythresh)
								hltpix++;
							totpix++;
						}
					}
				}
				hltlevel = (double)hltpix/(double)totpix - highlightpct;
				/* XXX doesn't work in case hue-var < 0 or hue+var > 1 */
				if (hltlevel > 0.0) {
					if (in[y][x].r > fringehue - fringevar
					&&  in[y][x].r < fringehue + fringevar) {
						hueaccuracy = fringevar - ABS(in[y][x].r - fringehue);
						//hueaccuracy=1;
						out[y][x].g /= ((1.0+hltlevel) * (1.0 + hueaccuracy))
								* bluereduction;
					}
				}
			}
		}
	}
	rgbarraytoxel(xsize,ysize,out,final,maxpix);
	pnm_writepnm(stdout,final,xsize,ysize,maxpix,format,0/*forceplain*/);
	pnm_freearray(in,ysize);
	pnm_freearray(out,ysize);
	exit(EXIT_SUCCESS);
	return 0;
}

