/*  PNMCONVOLVE - sharpen or blur image
 *
 * This file is part of PhotoPNMTools,
 * Release $Name: rel-1-3-0 $,
 * File rev. $Id: pnmconvolve.c,v 1.3 2003/02/22 12:21:18 schooten Exp $.
 *
 * Copyright (c) 2002, 2003 by B.W. van Schooten. All rights reserved.
 * This software is distributed under the Perl Artistic License.
 * No warranty. See file 'artistic.license' for more details.
 *
 * boris@13thmonkey.org
 * www.13thmonkey.org/~boris/photopnmtools/
 */

#include <pnm.h>
#include <stdio.h>
#include <math.h>

#include "stdpnm.h"

#define USAGE \
"pnmconvolve: convolution filter based sharpening and blurring.\n" \
"If you specify both sharpen and blur, the image will first be blurred, \n" \
"then sharpened.  If you enable the unsharp mask flag, then the image will\n"\
"not be blurred, but the blurred image will be subtracted from the original\n"\
"to form a sharpened image.\n" \
"Regular sharpening can be done effectively in HSY space: \n"\
"use ppmhsy -O, then use -B to operate on Y channel.\n" \
"For unsharp mask sharpening, you may require very high -s values,\n" \
"for example, a nice setting is  -u -s 10 -b 0.04 -i 2\n" \
"     -s [amount, 0.0..1.0 without -u, or likely > 1.0 with -u enabled]\n" \
"                                     : sharpen\n" \
"     -u                              : enable unsharp mask sharpening\n" \
"     -b [amount, 0.0..1.0]           : blur\n" \
"     -i [# of iterations, default 1] : blur multiple times\n" \
"     -R -G -B              : operate on R, G, or B only\n" \
"Sharpen parameters:\n" \
"     -T [softening threshold, 0.0..1.0, default 0.1]\n" \
"     -S [softening strength, 0.0..1.0, default 0.75]\n" \
"     -n [noise threshold, 0.0..1.0, default 0.05, N/A in unsharp mask mode]\n" \
"\n" \

#define ABS(a)  ((a) < 0 ? -(a) : (a) )

double sharpen = 0.0;
double blur = 0.0;
int bluriter=1;

int unsharpmask=0;

double soften_thresh=0.1, soften_strength=0.75;
double noise_thresh=0.05;

	double
sharpen_chan(double mid,  double tl, double tm, double tr,
   double ml,  double mr, double bl, double bm, double br) {
	double chan;
	// reduce noise amplification
	if (noise_thresh > 0.0) {
		if (ABS(tl - mid) < noise_thresh)
			tl -= (tl-mid)*(1.0 - ABS(tl-mid)/noise_thresh);
		if (ABS(tm - mid) < noise_thresh)
			tm -= (tm-mid)*(1.0 - ABS(tm-mid)/noise_thresh);
		if (ABS(tr - mid) < noise_thresh)
			tr -= (tr-mid)*(1.0 - ABS(tr-mid)/noise_thresh);
		if (ABS(ml - mid) < noise_thresh)
			ml -= (ml-mid)*(1.0 - ABS(ml-mid)/noise_thresh);
		if (ABS(mr - mid) < noise_thresh)
			mr -= (mr-mid)*(1.0 - ABS(mr-mid)/noise_thresh);
		if (ABS(bl - mid) < noise_thresh)
			bl -= (bl-mid)*(1.0 - ABS(bl-mid)/noise_thresh);
		if (ABS(bm - mid) < noise_thresh)
			bm -= (bm-mid)*(1.0 - ABS(bm-mid)/noise_thresh);
		if (ABS(br - mid) < noise_thresh)
			br -= (br-mid)*(1.0 - ABS(br-mid)/noise_thresh);
	}
	// convolve
	chan = (1.0 + sharpen*(4.0+4.0*0.7))*mid +
			- 0.7*sharpen*tl - sharpen*tm - 0.7*sharpen*tr
			-     sharpen*ml -                  sharpen*mr
			- 0.7*sharpen*bl - sharpen*bm - 0.7*sharpen*br;
	// soften large intensity changes
	if (chan - mid > soften_thresh) {
		chan -= (chan-mid - soften_thresh)*soften_strength;
	}
	if (chan - mid < -soften_thresh) {
		chan -= (chan-mid + soften_thresh)*soften_strength;
	}
	// clip
	//if (chan < 0.0) chan=0.0;
	//if (chan > maxpix) chan=maxpix;
	return chan;
}

	double
blur_chan(double mid,  double tl, double tm, double tr,
   double ml,  double mr, double bl, double bm, double br) {
	double chan;
	// convolve
	chan = (1.0 - blur*(4.0+4.0*0.7))*mid
			+ 0.7*blur*tl + blur*tm + 0.7*blur*tr
			+     blur*ml +               blur*mr
			+ 0.7*blur*bl + blur*bm + 0.7*blur*br;
	// clip
	//if (chan < 0.0) chan=0.0;
	//if (chan > maxpix) chan=maxpix;
	return chan;
}


	double
unsharp_chan(double orig, double blurred) {
	/* XXX we soften each channel separately.  Better to soften according to
	 * intensity change. */
	double chan;
	chan = (1.0 + sharpen)*orig - sharpen*blurred;
	// soften large intensity changes
	if (chan - orig > soften_thresh) {
		chan -= (chan-orig - soften_thresh)*soften_strength;
	}
	if (chan - orig < -soften_thresh) {
		chan -= (chan-orig + soften_thresh)*soften_strength;
	}
	return chan;
}



	int
main(int argc, char **argv) {
	xel **orig,**final;
	struct rgb **in,**out;

	/* options */
	int do_r=0,do_g=0,do_b=0;

	double r,g,b;
	char optch;

	pixval maxpix;
	int format;
	int xsize,ysize;
	int x,y;
	pnm_init(&argc,argv);
	while ( (optch = getopt(argc, argv, "s:b:i:uRGBT:S:n:")) != -1) {
		switch (optch) {
		case 's': sharpen=atof(optarg);
		break;
		case 'b': blur=atof(optarg);
		break;
		case 'i': bluriter=atoi(optarg);
		break;
		case 'u': unsharpmask=1;
		break;
		case 'R': do_r = 1;
		break;
		case 'G': do_g = 1;
		break;
		case 'B': do_b = 1;
		break;
		case 'T': soften_thresh=atof(optarg);
		break;
		case 'S': soften_strength=atof(optarg);
		break;
		case 'n': noise_thresh=atof(optarg);
		break;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc != 0) {
		fprintf(stderr,USAGE);
		exit(EXIT_FAILURE);
	}

	if ( unsharpmask && (blur == 0.0 || bluriter < 1 || sharpen==0.0 ) ) {
		fprintf(stderr,"ERROR: Bad unsharp mask operation\n\n");
		fprintf(stderr,USAGE);
		exit(EXIT_FAILURE);
	}
	if ( (sharpen == 0.0) 
	&&   (blur == 0.0    || bluriter < 1   ) ) {
		fprintf(stderr,"ERROR: No operation specified\n\n");
		fprintf(stderr,USAGE);
		exit(EXIT_FAILURE);
	}

	if (!do_r && !do_g && !do_b) {
		do_r=1;
		do_g=1;
		do_b=1;
	}

	blur /= 4 + 0.7*4;

	orig = pnm_readpnm(stdin,&xsize,&ysize,&maxpix,&format);
	in = rgballocarray(xsize, ysize, orig, maxpix);
	out = rgballocarray(xsize, ysize, NULL, maxpix);
	final = pnm_allocarray(xsize,ysize);

	/* XXX fix significance loss due to iterative operations on
	 * integers.  We avoid darkening by rouding properly (adding .5 before
	 * truncating) */

	while (blur > 0.0 && bluriter > 0) {
		bluriter--;
		for (y=0; y<ysize; y++) {
			for (x=0; x<xsize; x++) {
				out[y][x].r = in[y][x].r;
				out[y][x].g = in[y][x].g;
				out[y][x].b = in[y][x].b;
				if (x==0 || y==0 || x==xsize-1 || y==ysize-1) continue;
				if (do_r) {
					r = blur_chan(in[y][x].r,
							in[y-1][x-1].r, in[y-1][x].r, in[y-1][x+1].r,
							in[y  ][x-1].r,               in[y  ][x+1].r,
							in[y+1][x-1].r, in[y+1][x].r, in[y+1][x+1].r);
					out[y][x].r=r;
				}
				if (do_g) {
					g = blur_chan(in[y][x].g,
							in[y-1][x-1].g, in[y-1][x].g, in[y-1][x+1].g,
							in[y  ][x-1].g,               in[y  ][x+1].g,
							in[y+1][x-1].g, in[y+1][x].g, in[y+1][x+1].g);
					out[y][x].g=g;
				}
				if (do_b) {
					b = blur_chan(in[y][x].b,
							in[y-1][x-1].b, in[y-1][x].b, in[y-1][x+1].b,
							in[y  ][x-1].b,               in[y  ][x+1].b,
							in[y+1][x-1].b, in[y+1][x].b, in[y+1][x+1].b);
					out[y][x].b=b;
				}
			}
		}
		/* copy result back to in for next operation */
		for (y=0; y<ysize; y++) {
			for (x=0; x<xsize; x++) {
				in[y][x].r = out[y][x].r;
				in[y][x].g = out[y][x].g;
				in[y][x].b = out[y][x].b;
			}
		}
	}
	if (unsharpmask && sharpen > 0.0) {
		xelarraytorgb(xsize,ysize,orig,in,maxpix);
		/* in = original image, out = blurred image */
		for (y=0; y<ysize; y++) {
			for (x=0; x<xsize; x++) {
				if (do_r) {
					out[y][x].r = unsharp_chan(in[y][x].r,out[y][x].r);
				}
				if (do_g) {
					out[y][x].g = unsharp_chan(in[y][x].g,out[y][x].g);
				}
				if (do_b) {
					out[y][x].b = unsharp_chan(in[y][x].b,out[y][x].b);
				}
			}
		}
	} else if (!unsharpmask && sharpen > 0.0) {
		for (y=0; y<ysize; y++) {
			for (x=0; x<xsize; x++) {
				out[y][x].r = in[y][x].r;
				out[y][x].g = in[y][x].g;
				out[y][x].b = in[y][x].b;
				if (x==0 || y==0 || x==xsize-1 || y==ysize-1) continue;
				if (do_r) {
					r = sharpen_chan(in[y][x].r,
							in[y-1][x-1].r, in[y-1][x].r, in[y-1][x+1].r,
							in[y  ][x-1].r,               in[y  ][x+1].r,
							in[y+1][x-1].r, in[y+1][x].r, in[y+1][x+1].r);
					out[y][x].r=r;
				}
				if (do_g) {
					g = sharpen_chan(in[y][x].g,
							in[y-1][x-1].g, in[y-1][x].g, in[y-1][x+1].g,
							in[y  ][x-1].g,               in[y  ][x+1].g,
							in[y+1][x-1].g, in[y+1][x].g, in[y+1][x+1].g);
					out[y][x].g=g;
				}
				if (do_b) {
					b = sharpen_chan(in[y][x].b,
							in[y-1][x-1].b, in[y-1][x].b, in[y-1][x+1].b,
							in[y  ][x-1].b,               in[y  ][x+1].b,
							in[y+1][x-1].b, in[y+1][x].b, in[y+1][x+1].b);
					out[y][x].b=b;
				}
			}
		}
		/* copy result back to in for next operation *
		for (y=0; y<ysize; y++) {
			for (x=0; x<xsize; x++) {
				in[y][x].r = out[y][x].r;
				in[y][x].g = out[y][x].g;
				in[y][x].b = out[y][x].b;
			}
		}*/
	}
	rgbarraytoxel(xsize,ysize,out,final,maxpix);
	pnm_writepnm(stdout,final,xsize,ysize,maxpix,format,0/*forceplain*/);
	pnm_freearray(in,ysize);
	pnm_freearray(out,ysize);
	exit(EXIT_SUCCESS);
	return 0;
}

