/*  PNMDFT - Transform image between time and frequency domain
 *
 * This file is part of PhotoPNMTools,
 * Release $Name: rel-1-3-0 $,
 * File rev. $Id: pnmdft.c,v 1.2 2003/02/22 12:21:18 schooten Exp $.
 *
 * Copyright (c) 2002, 2003 by B.W. van Schooten. All rights reserved.
 * This software is distributed under the Perl Artistic License.
 * No warranty. See file 'artistic.license' for more details.
 *
 * boris@13thmonkey.org
 * www.13thmonkey.org/~boris/photopnmtools/
 */

#include <pnm.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#define USAGE \
"pnmdft: transform between the BLUE channel of a pnm and\n" \
"        a frequency domain map with\n" \
"        the red channel is the real part and the green channel the imaginary\n" \
"options: -r  reverse transform back to image, output is a pgm\n" \
"         -l  output values logarithmically (for viewing only)\n" \
"         -d  [destination depth in bytes, 1 or 2]\n" \
""

#define FFTSCALE 32765.0 

#ifndef FALSE
#define FALSE   (0)
#endif

#ifndef TRUE
#define TRUE    (!FALSE)
#endif

/* routine written by Paul Bourke,
 * taken from http://astronomy.swin.edu.au/~pbourke/analysis/fft2d/
 */

typedef struct {
	float real,imag;  /* should be double but we use float to save memory */
} COMPLEX;


	int
Powerof2(int size, int *log, int *power2) {
/*** returns 1 if size is power of 2, otherwise 0.
 *** Returns power2 = smallest power of 2
 *** that is bigger or equal to size, and log = log2(power2).
 ***/
	*log=0;
	*power2=1;
	while (*power2 < size) {
		*power2 <<= 1;
		(*log)++;
	}
	return (*power2 == size);
}

/*-------------------------------------------------------------------------
   This computes an in-place complex-to-complex FFT
   x and y are the real and imaginary arrays of 2^m points.
   dir =  1 gives forward transform
   dir = -1 gives reverse transform

     Formula: forward
                  N-1
                  ---
              1   \          - j k 2 pi n / N
      X(n) = ---   >   x(k) e                    = forward transform
              N   /                                n=0..N-1
                  ---
                  k=0

      Formula: reverse
                  N-1
                  ---
                  \          j k 2 pi n / N
      X(n) =       >   x(k) e                    = forward transform
                  /                                n=0..N-1
                  ---
                  k=0
*/
int FFT(int dir,int m,double *x,double *y)
{
   long nn,i,i1,j,k,i2,l,l1,l2;
   double c1,c2,tx,ty,t1,t2,u1,u2,z;

   /* Calculate the number of points */
   nn = 1;
   for (i=0;i<m;i++)
      nn *= 2;

   /* Do the bit reversal */
   i2 = nn >> 1;
   j = 0;
   for (i=0;i<nn-1;i++) {
      if (i < j) {
         tx = x[i];
         ty = y[i];
         x[i] = x[j];
         y[i] = y[j];
         x[j] = tx;
         y[j] = ty;
      }
      k = i2;
      while (k <= j) {
         j -= k;
         k >>= 1;
      }
      j += k;
   }

   /* Compute the FFT */
   c1 = -1.0;
   c2 = 0.0;
   l2 = 1;
   for (l=0;l<m;l++) {
      l1 = l2;
      l2 <<= 1;
      u1 = 1.0;
      u2 = 0.0;
      for (j=0;j<l1;j++) {
         for (i=j;i<nn;i+=l2) {
            i1 = i + l1;
            t1 = u1 * x[i1] - u2 * y[i1];
            t2 = u1 * y[i1] + u2 * x[i1];
            x[i1] = x[i] - t1;
            y[i1] = y[i] - t2;
            x[i] += t1;
            y[i] += t2;
         }
         z =  u1 * c1 - u2 * c2;
         u2 = u1 * c2 + u2 * c1;
         u1 = z;
      }
      c2 = sqrt((1.0 - c1) / 2.0);
      if (dir == 1)
         c2 = -c2;
      c1 = sqrt((1.0 + c1) / 2.0);
   }

   /* Scaling for forward transform */
   if (dir == 1) {
      for (i=0;i<nn;i++) {
         x[i] /= (double)nn;
         y[i] /= (double)nn;
      }
   }

   return(TRUE);
}

/*-------------------------------------------------------------------------
   Perform a 2D FFT inplace given a complex 2D array
   The direction dir, 1 for forward, -1 for reverse
   The size of the array (nx,ny)
   Return false if there are memory problems or
      the dimensions are not powers of 2
*/
int FFT2D(COMPLEX **c,int nx,int ny,int dir)
{
   int i,j;
   int m,twopm;
   double *real,*imag;

   /* Transform the rows */
   real = (double *)malloc(nx * sizeof(double));
   imag = (double *)malloc(nx * sizeof(double));
   if (real == NULL || imag == NULL)
      return(FALSE);
   if (!Powerof2(nx,&m,&twopm) || twopm != nx)
      return(FALSE);
   for (j=0;j<ny;j++) {
      for (i=0;i<nx;i++) {
         real[i] = c[i][j].real;
         imag[i] = c[i][j].imag;
      }
      FFT(dir,m,real,imag);
      for (i=0;i<nx;i++) {
         c[i][j].real = real[i];
         c[i][j].imag = imag[i];
      }
   }
   free(real);
   free(imag);

   /* Transform the columns */
   real = (double *)malloc(ny * sizeof(double));
   imag = (double *)malloc(ny * sizeof(double));
   if (real == NULL || imag == NULL)
      return(FALSE);
   if (!Powerof2(ny,&m,&twopm) || twopm != ny)
      return(FALSE);
   for (i=0;i<nx;i++) {
      for (j=0;j<ny;j++) {
         real[j] = c[i][j].real;
         imag[j] = c[i][j].imag;
      }
      FFT(dir,m,real,imag);
      for (j=0;j<ny;j++) {
         c[i][j].real = real[j];
         c[i][j].imag = imag[j];
      }
   }
   free(real);
   free(imag);

   return(TRUE);
}




	void
clear_matrix(COMPLEX **buf,int xsize,int ysize) {
	int x,y;
	for (y=0; y<ysize; y++) {
		for (x=0; x<xsize; x++) {
			buf[x][y].real=0.0;
			buf[x][y].imag=0.0;
		}
	}
}


/* 14 for 8 bits */
/* 2800 for 16 bits */
/*#define LOGEXP 1800.0*/

	double
get_logexp(pixval maxpix) {
	if (maxpix == 255) {
		return 14.0;
	}
	if (maxpix == 65535) {
		return 1800.0;
	}
	fprintf(stderr,"Logarithmic transform requires depth of 255 or 65535\n");
	exit(EXIT_FAILURE);
}


	long
to_log(double value, long range, double logexp) {
//r = 255.0 + LOGEXP*log(fabs(buf[x][y].real));
	int result;
	if (value > 0.000000001) {
		result = rint( range + logexp*log( value) );
		if (result < (range>>1)) result= range>>1;
		if (result >= range) fprintf(stderr,"Result overflow!!\n");
	} else if (value < -0.000000001) {
		result = rint(       - logexp*log(-value) );
		if (result >= (range>>1)) result = (range>>1) - 1;
		if (result < 0) fprintf(stderr,"Result underflow!!\n");
	} else { /* value = 0.0 */
		return range>>1;
	}
	return result;
}

	double
from_log(long log, long range, double logexp) {
	if (log >= (range>>1)) {
		return  exp( ( (float)log - (float)range)/logexp );
	} else {
		return -exp( (-(float)log               )/logexp );
	}
}


	int
main(int argc, char **argv) {
	xel **in,**out;
	COMPLEX **buf;

	int direction=1; /* 1=forward -1=reverse */
	int logarithmic=0;

	pixval maxpix;
	int format;
	int xsize,ysize;

	int fftxsize,fftysize;

	int x,y;
	int logpower;

	int destdepth=2;
	pixval destmaxpix;
	double logexp;

	char optch;

	double r,g,b;

	double real,imag;

	//if (argc == 1) {
	//	fprintf(stderr,USAGE);
	//	exit(EXIT_FAILURE);
	//}
	while ( (optch = getopt(argc, argv, "rld:")) != -1) {
		switch (optch) {
		case 'r': direction=-1;
		break;
		case 'l': logarithmic=1;
		break;
		case 'd': destdepth= atoi(optarg);
		break;
		}
	}
	if (destdepth==1) { destmaxpix=255; }
	else if (destdepth==2) { destmaxpix=65535; }
	else {
		fprintf(stderr,"Bad destination depth, must be 1 or 2 bytes\n");
		exit(EXIT_FAILURE);
	}

	pnm_init(&argc,argv);
	in = pnm_readpnm(stdin,&xsize,&ysize,&maxpix,&format);

	Powerof2(xsize,&logpower,&fftxsize);
	Powerof2(ysize,&logpower,&fftysize);

	out = pnm_allocarray(fftxsize,fftysize);

	buf = malloc(sizeof(COMPLEX *)*fftxsize);
	if (!buf) exit(EXIT_FAILURE);
	for (x=0; x<fftxsize; x++) {
		buf[x] = malloc(sizeof(COMPLEX)*fftysize);
		if (!buf[x]) exit(EXIT_FAILURE);
	}

	if (direction == 1) {
		/* do blue channel */
		clear_matrix(buf,fftxsize,fftysize);
		for (y=0; y<ysize; y++) {
			for (x=0; x<xsize; x++) {
				buf[x][y].real = (float)in[y][x].b / (float)maxpix;
			}
		}
		if (FFT2D(buf,fftxsize,fftysize,direction) == FALSE) {
			fprintf(stderr,"Error in FFT!\n");
			exit(EXIT_FAILURE);
		}
		// 372,224, 37x37
//		for (x=180; x<180+35; x++) {
//			for (y=108; y<108+35; y++) {
//				buf[x][y].real = 0.0;
//				buf[x][y].imag = 0.0;
//			}
//		}
//		for (x=3886; x<3886+35; x++) {
//			for (y=1914; y<1914+35; y++) {
//				buf[x][y].real = 0.0;
//				buf[x][y].imag = 0.0;
//			}
//		}
//		for (x=3886; x<3886+35; x++) {
//			for (y=92; y<92+35; y++) {
//				buf[x][y].real = 0.0;
//				buf[x][y].imag = 0.0;
//			}
//		}
//		for (x=180; x<180+35; x++) {
//			for (y=1914; y<1914+35; y++) {
//				buf[x][y].real = 0.0;
//				buf[x][y].imag = 0.0;
//			}
//		}
		if (logarithmic) {
			logexp = get_logexp(destmaxpix);
			for (y=0; y<fftysize; y++) {
				for (x=0; x<fftxsize; x++) {
					r = to_log(buf[x][y].real,destmaxpix,logexp);
					g = to_log(buf[x][y].imag,destmaxpix,logexp);
					//r = 255.0 + 20*log(fabs(buf[x][y].real));
					//if (r < 0.0) r=0.0;
					//if (r > 255.0) r=255.0;
					out[y][x].r = r;
					//g = 255.0 + 20*log(fabs(buf[x][y].imag));
					//if (g < 0.0) g=0.0;
					//if (g > 255.0) g=255.0;
					out[y][x].g = g;
					//fprintf(stderr,"%f %f %f %f\n",
					//		fabs(buf[x][y].real),fabs(buf[x][y].imag),r,g);
				}
			}
			pnm_writepnm(stdout,out,fftxsize,fftysize,destmaxpix,
				PPM_FORMAT,0/*forceplain*/);
		} else {
			for (y=0; y<fftysize; y++) {
				for (x=0; x<fftxsize; x++) {
					out[y][x].r = 32768 + rint(FFTSCALE * buf[x][y].real);
					out[y][x].g = 32768 + rint(FFTSCALE * buf[x][y].imag);
				}
			}
			pnm_writepnm(stdout,out,fftxsize,fftysize,65535,
				PPM_FORMAT,0/*forceplain*/);
		}
	} else {
		/* red/green = real/imag channel */
		clear_matrix(buf,fftxsize,fftysize);
		if (logarithmic) {
			logexp = get_logexp(maxpix);
			for (y=0; y<ysize; y++) {
				for (x=0; x<xsize; x++) {
					buf[x][y].real = from_log(in[y][x].r,maxpix,logexp);
					buf[x][y].imag = from_log(in[y][x].g,maxpix,logexp);
				}
			}
		} else {
			for (y=0; y<ysize; y++) {
				for (x=0; x<xsize; x++) {
					buf[x][y].real = ((float)in[y][x].r - 32768) / FFTSCALE;
					buf[x][y].imag = ((float)in[y][x].g - 32768) / FFTSCALE;
				}
			}
		}
		if (FFT2D(buf,fftxsize,fftysize,direction) == FALSE) {
			fprintf(stderr,"Error in FFT!\n");
			exit(EXIT_FAILURE);
		}
		for (y=0; y<fftysize; y++) {
			for (x=0; x<fftxsize; x++) {
				r = 255.0 * buf[x][y].real;
				if (r < 0.0) r=0;
				if (r > 255.0) r=255.0;
				//g = 255.0 * buf[x][y].imag;
				//if (g < 0.0) g=0;
				//if (g > 255.0) g=255.0;
				//out[y][x].g = rint(g);

				out[y][x].r = rint(r);
				out[y][x].g = rint(r);
				out[y][x].b = rint(r);
			}
		}
		pnm_writepnm(stdout,out,fftxsize,fftysize,255,
				PGM_FORMAT,0/*forceplain*/);
	}

	pnm_freearray(in,ysize);
	pnm_freearray(out,fftysize);
	exit(EXIT_SUCCESS);
	return 0;
}



