/*  PNMDIFF - calculate difference between images
 *
 * This file is part of PhotoPNMTools,
 * Release $Name: rel-1-3-0 $,
 * File rev. $Id: pnmdiff.c,v 1.3 2003/03/06 10:07:47 schooten Exp $.
 *
 * Copyright (c) 2002, 2003 by B.W. van Schooten. All rights reserved.
 * This software is distributed under the Perl Artistic License.
 * No warranty. See file 'artistic.license' for more details.
 *
 * boris@13thmonkey.org
 * www.13thmonkey.org/~boris/photopnmtools/
 */

#include <pnm.h>
#include <stdio.h>
#include <math.h>

#define MAX(a,b) ( (a) > (b) ? (a) : (b) )

#define USAGE \
"pnmdiff: calculate difference between two pnms.\n" \
"Expected arguments:   <pnm>  <pnm to subtract>\n" \
"options: -e   calculate error image (dst = maxpix/2 + src1 - src2)\n"\
"         -s   subtract (dst = src1 - src2)\n"\
"         -d   divide (dst = (src1-maxpix/2) / (src2+1) + maxpix/2 )\n"\
"         -f [clip floor value, an int between 0 and 65535, default=0]\n" \
"         -c [clip ceiling value, an int between 0 and 65535, default=maxpix]\n"


enum {Errorimage, Subtract, Divide};

	int
main(int argc, char **argv) {
	xel **in1,**in2,**out;

	FILE *infile1,*infile2;

	int mode=Errorimage;
	long clipfloor=0,clipceil=-1;
	long r,g,b;
	char optch;

	pixval maxpix;
	int format;
	int xsize1,ysize1,xsize2,ysize2;
	int dstxsize,dstysize;
	int x,y;
	pnm_init(&argc,argv);
	while ( (optch = getopt(argc, argv, "esdf:c:")) != -1) {
		switch (optch) {
		case 'e': mode=Errorimage;
		break;
		case 's': mode=Subtract;
		break;
		case 'd': mode=Divide;
		break;
		case 'f': clipfloor=atol(optarg);
		break;
		case 'c': clipceil=atol(optarg);
		break;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc != 2) {
		fprintf(stderr,USAGE);
		exit(EXIT_FAILURE);
	}

	infile1 = fopen(argv[0],"rb");
	infile2 = fopen(argv[1],"rb");
	in1 = pnm_readpnm(infile1,&xsize1,&ysize1,&maxpix,&format);
	in2 = pnm_readpnm(infile2,&xsize2,&ysize2,&maxpix,&format);
	fclose(infile1);
	fclose(infile2);
	dstxsize = MAX(xsize1,xsize2);
	dstysize = MAX(ysize1,ysize2);
	out = pnm_allocarray(dstxsize,dstysize);

	if (clipceil < 0) clipceil  = maxpix;
	for (y=0; y<dstysize; y++) {
		for (x=0; x<dstxsize; x++) {
			if (mode == Errorimage) {
				out[y][x].r = maxpix>>1;
				out[y][x].g = maxpix>>1;
				out[y][x].b = maxpix>>1;
			} else if (mode == Subtract || mode == Divide) {
				out[y][x].r = 0;
				out[y][x].g = 0;
				out[y][x].b = 0;
			}
			if (x < xsize1 && y < ysize1) {
				out[y][x].r += in1[y][x].r;
				out[y][x].g += in1[y][x].g;
				out[y][x].b += in1[y][x].b;
			}
			if (x < xsize2 && y < ysize2) {
				if (mode==Subtract || mode==Errorimage) {
					r = out[y][x].r - in2[y][x].r;
					g = out[y][x].g - in2[y][x].g;
					b = out[y][x].b - in2[y][x].b;
					if (r<clipfloor) r=clipfloor;
					if (r>clipceil) r=clipceil;
					if (g<clipfloor) g=clipfloor;
					if (g>clipceil) g=clipceil;
					if (b<clipfloor) b=clipfloor;
					if (b>clipceil) b=clipceil;
					out[y][x].r = r;
					out[y][x].g = g;
					out[y][x].b = b;
				} else if (mode==Divide) {
					out[y][x].r = (out[y][x].r - (maxpix>>1))
							/ (in2[y][x].r+1) + (maxpix>>1);
					out[y][x].g = (out[y][x].g - (maxpix>>1))
							/ (in2[y][x].g+1) + (maxpix>>1);
					out[y][x].b = (out[y][x].b - (maxpix>>1))
							/ (in2[y][x].b+1) + (maxpix>>1);
				}
			}
		}
	}

	pnm_writepnm(stdout,out,dstxsize,dstysize,maxpix,format,0/*forceplain*/);
	pnm_freearray(in1,ysize1);
	pnm_freearray(in2,ysize2);
	pnm_freearray(out,dstysize);
	exit(EXIT_SUCCESS);
	return 0;
}

