/*  PNMRESAMPLE - shrink and expand images
 *
 * This file is part of PhotoPNMTools,
 * Release $Name: rel-1-3-0 $,
 * File rev. $Id: pnmresample.c,v 1.7 2003/04/09 00:13:46 schooten Exp $.
 *
 * Copyright (c) 2002, 2003 by B.W. van Schooten. All rights reserved.
 * This software is distributed under the Perl Artistic License.
 * No warranty. See file 'artistic.license' for more details.
 *
 * boris@13thmonkey.org
 * www.13thmonkey.org/~boris/photopnmtools/
 */

#include <pnm.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

/* sinc window is a float value */
/* 1 = XV-type enlargement (nice for black&white pictures)
 * 2 = strong sharpening effect
 * 3 = less sharpening
 * 4 = almost as strong as 2
 * 6 = little difference with 4.
 * 
 * In general:
 * even => strong sharpening
 * odd  => less sharpening
 * */

#define USAGE \
"pnmresize: resize a pnm from stdin to stdout\n" \
"options: -x [width]\n" \
"         -y [height]\n" \
"         -X [horiz scale factor (default: same as vert)]\n" \
"         -Y [vert scale factor (default: same as horiz)]\n" \
"         -b force square aspect, using given size as bounding box\n" \
"         -B ignore (do not weigh) black pixels when shrinking\n" \
"         -d disable anti-aliasing (i.e. use sync when shrinking)\n" \
"enlarging only:\n" \
"         -w [sync window size (default 4)]\n" \
"         -f [sync fade interval (default 2)]\n" \
"         Useful values:\n" \
"         -w 1 -f 0   close to bilinear (for line drawings)\n" \
"         -w 2 -f 0.5 strong sharpening effect\n" \
"         -w 3 -f 1   some sharpening effect (photos)\n" \
"         -w 4 -f 2   good-looking and reasonable approximation to sinc (photos)\n" \
"shrinking only:\n" \
"         -s [weighting slope (0=linear, 1=sine, 2..4=tanh, 5=step, default:0)]\n" \
"         step (5):     lose little sharpness but some aliasing\n" \
"         linear (0):   softening effect but aliasing much reduced\n" \
"         curve (1..4): curves with increasing steepness,\n" \
"                       compromise between step and linear\n"

/* must be power of 2 */
#define BUFMODULO      4
#define BUFIDX(x,y)     ( (int)(x) + dstxsize*((int)(y)&(BUFMODULO-1)) )



float sinc_window=4, fade_window=2;

int weighting_slope=0;

int antialiasing=1;

	float
radius(float scale) {
	if (scale > 2) {
		return scale / 1.0;
	} else {
		return 2.01;
	}
}

	float
getweight(float diffx,float diffy,float scalex,float scaley) {
	float wght;
	wght= (0.5*radius(scalex)-diffx)+(0.5*radius(scaley)-diffy);
	if (wght < 0) wght=0;
	return wght;
}

struct pixval {
	float r,g,b;
	float weight;
};

	void
add_rgb(struct pixval *buf, float weight, xel *in) {
	buf->r += in->r*weight;
	buf->g += in->g*weight;
	buf->b += in->b*weight;
	buf->weight += weight;
}
/* macro version for speed */
#define ADD_RGB(buf,Weight,in) do { \
	(buf)->r += (in)->r*(Weight); \
	(buf)->g += (in)->g*(Weight); \
	(buf)->b += (in)->b*(Weight); \
	(buf)->weight += (Weight); } while (0);


	inline float
sinc(float x) {
	if (x > -0.0001 && x < 0.0001) return 1;
	if (x < -sinc_window+fade_window) {
		return sin(M_PI*x) / (M_PI*x) * ((x + sinc_window)/(float)fade_window);
	} else if (x > sinc_window-fade_window) {
		return sin(M_PI*x) / (M_PI*x) * ((sinc_window - x)/(float)fade_window);
	} else {
		return sin(M_PI*x) / (M_PI*x);
	}
}

	inline float
mexican_hat(float x,float y) {
	float dist;
	int x_near_0 = x > -0.0001 && x < 0.0001;
	int y_near_0 = y > -0.0001 && y < 0.0001;
	if (x_near_0 && y_near_0) return 1;
	dist = M_PI*sqrt(x*x+y*y);
	if (dist > M_PI*sinc_window) return 0;
	return sin(dist) / (dist);
}


	int
main(int argc, char **argv) {
	xel **in,**out;
	struct pixval *buf;

	pixval maxpix;
	int format;
	int srcxsize,srcysize;
	int dstxsize=0,dstysize=0; /* 0 indicates not yet defined */
	float xscale=0.0,yscale=0.0;
	/* 0.0 indicates not yet defined => use absolute dstxsize,dstysize */

	int squareasp=0;
	int no_weigh_black=0;

	float xfrac,yfrac;
	float weight, xweight,yweight;

	float dx,dy;

	float sincy;
	xel *thisxel;

	float srcx,srcy;
	float scalex,scaley;
	int dstx,dsty;
	int sx,sy;

	char optch;

	int underflows,overflows,neg_weights;

	float r,g,b, wght;
	float weighttotal;

	int floordx,floordy;
	xel *in_sy_sx;
	struct pixval *bufelem;

	if (argc == 1) {
		fprintf(stderr,USAGE);
		exit(EXIT_FAILURE);
	}
	while ( (optch = getopt(argc, argv, "x:X:y:Y:w:f:s:o:bBd")) != -1) {
		switch (optch) {
		case 'x': dstxsize = atoi(optarg);
		break;
		case 'y': dstysize = atoi(optarg);
		break;
		case 'X': xscale = atof(optarg);
		break;
		case 'Y': yscale = atof(optarg);
		break;
		case 'w': sinc_window= atof(optarg);
		break;
		case 'f': fade_window= atof(optarg);
		break;
		case 's': weighting_slope= atoi(optarg);
		break;
		case 'b': squareasp=1;
		break;
		case 'B': no_weigh_black=1;
		break;
		case 'd': antialiasing=0;
		break;
		}
	}

	pnm_init(&argc,argv);
	fprintf(stderr,"READ FILE\n");
	in = pnm_readpnm(stdin,&srcxsize,&srcysize,&maxpix,&format);
	if (xscale == 0.0 && yscale==0.0 && dstxsize==0 && dstysize==0) {
		fprintf(stderr,"No target size or scale factor supplied\n");
		exit(EXIT_FAILURE);
	}
	fprintf(stderr,"PROCESS\n");
	/* get dstsize from scale if applicable */
	if (xscale != 0.0) {
		dstxsize = srcxsize * xscale;
	}
	if (yscale != 0.0) {
		dstysize = srcysize * yscale;
	}
	/* ensure square aspect if applicable (i.e. both sizes are defined) */
	if (squareasp && dstxsize != 0 && dstysize != 0) {
		xscale = (float)dstxsize / (float)srcxsize;
		yscale = (float)dstysize / (float)srcysize;
		if (xscale > yscale) dstxsize = srcxsize * yscale;
		else                 dstysize = srcysize * xscale;
	}
	/* ensure both x and y sizes are defined */
	if (dstxsize == 0) {
		if (yscale == 0.0) {
			dstxsize = (float)srcxsize * ((float)dstysize / (float)srcysize);
		} else {
			dstxsize = (float)srcxsize * yscale;
		}
	}
	if (dstysize == 0) {
		if (xscale == 0.0) {
			dstysize = (float)srcysize * ((float)dstxsize / (float)srcxsize);
		} else {
			dstysize = (float)srcysize * xscale;
		}
	}
	fprintf(stderr,"New size: %d %d\n",dstxsize,dstysize);
	out = pnm_allocarray(dstxsize,dstysize);


	if (dstxsize < srcxsize && dstysize < srcysize && antialiasing) {

		/* if we scale down, we use source-driven resampling.  Each source
		 * pixel has an influence on the four destination pixels that are
		 * closest to it.  The influence is weighted according to the distance
		 * it has from each pixel. The distance measure we use is xdistance +
		 * ydistance.  The distance is translated to a weight function using a
		 * sinc-like curve.  The simplest curve is a triangle with its maximum
		 * value at 0.
		 */

		/* The algorithm tries to maintain some desirable properties of
		 * downsampling:
		 *
		 * * average brightness of thin lines remains equal everywhere
		 *
		 * * the picture loses as little sharpness as possible
		 *
		 * * aliasing effects (such as found when downsampling high-contrast
		 * textures) should be minimised
		 *
		 */

		buf = malloc(sizeof(struct pixval)*dstxsize*BUFMODULO);

		for (dsty=0; dsty<dstysize; dsty++) {
			for (dstx=0; dstx<dstxsize; dstx++) {
				buf[BUFIDX(dstx,dsty)].r = 0.0;
				buf[BUFIDX(dstx,dsty)].g = 0.0;
				buf[BUFIDX(dstx,dsty)].b = 0.0;
				buf[BUFIDX(dstx,dsty)].weight = 0.0;
				out[dsty][dstx].r=0;
				out[dsty][dstx].g=0;
				out[dsty][dstx].b=0;
			}
		}

		scalex = (float)dstxsize / (float)srcxsize;
		scaley = (float)dstysize / (float)srcysize;
	
		for (sy=0, dy=0.0; sy<srcysize; sy++, dy+=scaley) {
			yfrac = dy - floor(dy);
			switch (weighting_slope) {
				case 5: yweight = yfrac < 0.5 ? 0 : 1;
				break;
				case 4: yweight = 0.5 + 0.5*tanh(15*(yfrac-0.5));
				break;
				case 3: yweight = 0.5 + 0.5*tanh(8*(yfrac-0.5));
				break;
				case 2: yweight = 0.5 + 0.5*tanh(5*(yfrac-0.5));
				break;
				case 1: yweight = 0.5 - 0.5*cos(yfrac*M_PI);
				break;
				case 0: yweight = yfrac;
				break;
				default : fprintf(stderr,"illegal weighting slope\n");
					yweight = yfrac;
			}
			for (sx=0, dx=0.0; sx<srcxsize; sx++, dx+=scalex) {
				xfrac = dx - floor(dx);
				switch (weighting_slope) {
					case 5: xweight = xfrac < 0.5 ? 0 : 1;
					break;
					case 4: xweight = 0.5 + 0.5*tanh(15*(xfrac-0.5));
					break;
					case 3: xweight = 0.5 + 0.5*tanh(8*(xfrac-0.5));
					break;
					case 2: xweight = 0.5 + 0.5*tanh(5*(xfrac-0.5));
					break;
					case 1: xweight = 0.5 - 0.5*cos(xfrac*M_PI);
						/*almost same as tanh(4*x)*/
					break;
					case 0: xweight = xfrac;
					break;
					default : fprintf(stderr,"illegal weighting slope\n");
						xweight = xfrac;
				}
				floordx = floor(dx);
				floordy = floor(dy);
				in_sy_sx = &(in[sy][sx]);
				if (no_weigh_black)
					if (in_sy_sx->r==0 && in_sy_sx->g==0 && in_sy_sx->b==0)
						continue;
				bufelem = &(buf[BUFIDX(floordx,floordy)]);
				ADD_RGB(bufelem,
						(1.0-xweight)*(1.0-yweight),
						in_sy_sx);
				if (dx<dstxsize-1) {
					bufelem++;
					//bufelem = &(buf[BUFIDX(floordx+1,floordy)]);
					ADD_RGB(bufelem,
						xweight*(1.0-yweight),
						in_sy_sx);
				}
				if (dy<dstysize-1) {
					bufelem = &(buf[BUFIDX(floordx,floordy+1)]);
					ADD_RGB(bufelem,
						(1.0-xweight)*yweight,
						in_sy_sx);
					if (dx<dstxsize-1) {
						bufelem++;
						//bufelem = &(buf[BUFIDX(floordx+1,floordy+1)]);
						ADD_RGB(bufelem,
							xweight*yweight,
							in_sy_sx);
					}
				}
			}
			if (floor(dy + scaley) > floor(dy)) {
				/* line finished -> write to out */
				dsty = floor(dy);
				for (dstx=0; dstx<dstxsize; dstx++) {
					weight = buf[BUFIDX(dstx,dsty)].weight;
					if (weight!=0.0) {
						out[dsty][dstx].r = rint(buf[BUFIDX(dstx,dsty)].r/weight);
						out[dsty][dstx].g = rint(buf[BUFIDX(dstx,dsty)].g/weight);
						out[dsty][dstx].b = rint(buf[BUFIDX(dstx,dsty)].b/weight);
					}
					buf[BUFIDX(dstx,dsty)].r = 0.0;
					buf[BUFIDX(dstx,dsty)].g = 0.0;
					buf[BUFIDX(dstx,dsty)].b = 0.0;
					buf[BUFIDX(dstx,dsty)].weight = 0.0;
				}
			}
		}

	} else {

		/* If we scale up, we use destination-driven resampling.  Each
		 * destination pixel value is interpolated from some of its
		 * neighbouring source pixels.
		 */
	
		/* sinc function: sin(pi*x) / pi*x   (with x=0 -> sinc(x)=1) */

		underflows=0;
		overflows=0;
		neg_weights=0;

		scalex = (float)srcxsize / (float)dstxsize;
		scaley = (float)srcysize / (float)dstysize;
	
		for (dsty=0,srcy=0.0; dsty<dstysize; dsty++, srcy+=scaley) {
			if (dsty%10) fprintf(stderr,"."); else fprintf(stderr,"+");
			for (dstx=0,srcx=0.0; dstx<dstxsize; dstx++, srcx+=scalex) {
				r=0.0;
				g=0.0;
				b=0.0;
				weighttotal=0.0;
				for (sy=ceil(srcy-sinc_window); sy<=floor(srcy+sinc_window); sy++) {
					if (sy < 0) continue;
					if (sy >= srcysize) break;
					sincy = sinc(sy-srcy);
					for (sx=ceil(srcx-sinc_window); sx<=floor(srcx+sinc_window); sx++) {
						if (sx < 0) continue;
						if (sx >= srcxsize) break;
						weight = sincy*sinc(sx-srcx);
						//weight = mexican_hat(sx-srcx,sy-srcy);
						weighttotal += weight;
						thisxel = &in[sy][sx];
						r += weight * thisxel->r;
						g += weight * thisxel->g;
						b += weight * thisxel->b;
					}
				}
				if (weighttotal < 0) neg_weights++;
				if (r<0.0 || g<0.0 || b<0.0) underflows++;
				if (r>=255.5 || g>=255.5 || b>=255.5) overflows++;
				r /= weighttotal;
				g /= weighttotal;
				b /= weighttotal;
				if (r>255.0) r = 255.0;
				if (g>255.0) g = 255.0;
				if (b>255.0) b = 255.0;
				if (r<0.0) r = 0.0;
				if (g<0.0) g = 0.0;
				if (b<0.0) b = 0.0;
				out[dsty][dstx].r = rint(r);
				out[dsty][dstx].g = rint(g);
				out[dsty][dstx].b = rint(b);
			}
		}
		fprintf(stderr,
				"\noverflows: %d, underflows: %d, negative weights: %d\n",
				overflows, underflows, neg_weights);

#if 0
				for (sy=ceil(srcy-radius(scaley));
				sy<=floor(srcy+radius(scaley));
				sy++) {
					for (sx=ceil(srcx-radius(scalex));
					sx<=floor(srcx+radius(scalex));
					sx++) {
						wght =
							getweight(fabsf(srcx-sx),fabsf(srcy-sy),scalex,scaley);
						if (sx >= 0 && sy >= 0 
						&& sy < srcysize && sx < srcxsize) {
							r += in[sy][sx].r*wght;
							g += in[sy][sx].g*wght;
							b += in[sy][sx].b*wght;
							weighttotal+=wght;
						} /* else black */
					}
				}
				out[dsty][dstx].r = r/weighttotal;
				out[dsty][dstx].g = g/weighttotal;
				out[dsty][dstx].b = b/weighttotal;
				if (out[dsty][dstx].r > maxpix 
				||  out[dsty][dstx].g > maxpix 
				||  out[dsty][dstx].b > maxpix) {
					fprintf(stderr,"PAST MAXPIX");
				}
			}
		}
#endif

	}

	fprintf(stderr,"WRITE FILE\n");
	pnm_writepnm(stdout,out,dstxsize,dstysize,maxpix,format,0/*forceplain*/);
	pnm_freearray(in,srcysize);
	pnm_freearray(out,dstysize);
	exit(EXIT_SUCCESS);
	return 0;
}

