/*  PPMLEVELS - levels tool for renormalising brightness
 *
 * This file is part of PhotoPNMTools,
 * Release $Name: rel-1-3-0 $,
 * File rev. $Id: ppmlevels.c,v 1.4 2003/03/30 23:43:41 schooten Exp $.
 *
 * Copyright (c) 2002, 2003 by B.W. van Schooten. All rights reserved.
 * This software is distributed under the Perl Artistic License.
 * No warranty. See file 'artistic.license' for more details.
 *
 * boris@13thmonkey.org
 * www.13thmonkey.org/~boris/photopnmtools/
 */

#include <pnm.h>
#include <stdio.h>
#include <math.h>

#include "stdpnm.h"

#define MAX(a,b) ( (a) > (b) ? (a) : (b) )

#define USAGE \
"ppmlevels: levels tool for renormalising brightness\n"\
"    $Id: ppmlevels.c,v 1.4 2003/03/30 23:43:41 schooten Exp $\n"\
"    (C) 2002, 2003 Boris van Schooten   boris@13thmonkey.org\n" \
"options: -m : manual: disable auto levels, use top and bottom values as\n"\
"              absolute values rather than percentiles.\n"\
"         -0 -1 -2 : disable red, green, blue operations respectively\n"\
"         -h [nr of bars] : output histogram to stdout\n"\
"         -r [bottom red   level, 0.0..1.0, default=0.01]\n"\
"         -g [bottom green level, 0.0..1.0, default=0.01]\n"\
"         -b [bottom blue  level, 0.0..1.0, default=0.01]\n"\
"         -R [top    red   level, 0.0..1.0, default=0.995]\n"\
"         -G [top    green level, 0.0..1.0, default=0.995]\n"\
"         -B [top    blue  level, 0.0..1.0, default=0.995]\n"\
"\n"


#define R 0
#define G 1
#define B 2

char *colname[] = {"red","green","blue"};

double devicegamma=0.5;

	int
main(int argc, char **argv) {
	xel **in,**out=NULL;

	double botlevel[3] = {0.01,0.01,0.01}, toplevel[3]={0.995,0.995,0.995};
	int histlevels=250;
	int outputhist=0;
	int automode=1;
	int enable[3] = {1,1,1};
	double r,g,b;
	char optch;

	long *hist[3];

	pixval maxpix;
	int format;
	int xsize,ysize;
	int x,y, i,c;

	int totpix=0, curpix;
	double botpct[3],toppct[3];

	stdpnminit();
	pnm_init(&argc,argv);
	while ( (optch = getopt(argc, argv, "m012h:r:g:b:R:G:B:")) != -1) {
		switch (optch) {
		case 'm': automode=0;
		break;
		case '0': enable[0]=0;
		break;
		case '1': enable[1]=0;
		break;
		case '2': enable[2]=0;
		break;
		case 'h': outputhist=1; histlevels=atoi(optarg);
		break;
		case 'r': botlevel[R]=atof(optarg);
		break;
		case 'g': botlevel[G]=atof(optarg);
		break;
		case 'b': botlevel[B]=atof(optarg);
		break;
		case 'R': toplevel[R]=atof(optarg);
		break;
		case 'G': toplevel[G]=atof(optarg);
		break;
		case 'B': toplevel[B]=atof(optarg);
		break;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc != 0) {
		fprintf(stderr,USAGE);
		exit(EXIT_FAILURE);
	}

	in = pnm_readpnm(stdin,&xsize,&ysize,&maxpix,&format);
	if (!outputhist) {
		out = pnm_allocarray(xsize,ysize);
	}
	hist[R] = malloc(sizeof(long)*histlevels);
	hist[G] = malloc(sizeof(long)*histlevels);
	hist[B] = malloc(sizeof(long)*histlevels);
	for (y=0; y<ysize; y++) {
		for (x=0; x<xsize; x++) {
			hist[R][(in[y][x].r*(histlevels-1) / maxpix)]++;
			hist[G][(in[y][x].g*(histlevels-1) / maxpix)]++;
			hist[B][(in[y][x].b*(histlevels-1) / maxpix)]++;
			totpix++;
		}
	}
	/* in automode, we convert the levels to percentiles */
	if (automode) {
		for (c=0; c<3; c++) {
			curpix=0;
			for (i=0; i<histlevels; i++) {
				curpix += hist[c][i];
				if (curpix > (totpix*botlevel[c])) {
					botpct[c] = (double)i / histlevels;
					break;
				}
			}
			curpix=0;
			for (i=0; i<histlevels; i++) {
				curpix += hist[c][i];
				if (curpix > (totpix*toplevel[c])) {
					toppct[c] = (double)i / histlevels;
					break;
				}
			}
		}
		for (c=0; c<3; c++) {
			botlevel[c] = botpct[c];
			toplevel[c] = toppct[c];
		}
	}
	/*fprintf(stderr,"tr%f br%f tg%f bg%f\n",
			toppct[R], botpct[G],
			toppct[G], botpct[G]);*/
	/* gamma correction levels */
	for (c=0; c<3; c++) {
		toplevel[c] = cierectolinear(toplevel[c],c);
		botlevel[c] = cierectolinear(botlevel[c],c);
	}
	if (!outputhist && automode) {
		for (y=0; y<ysize; y++) {
			for (x=0; x<xsize; x++) {
				if (enable[R]) {
					r = (double)in[y][x].r / maxpix;
					r = cierectolinear(r,0);
					r = (r - botlevel[R]) / (toplevel[R] - botlevel[R]);
					if (r<0) r = 0;
					if (r>1) r = 1;
					r = lineartocierec(r,0);
					out[y][x].r = r*maxpix;
				} else {
					out[y][x].r = in[y][x].r;
				}
				if (enable[G]) {
					g = (double)in[y][x].g / maxpix;
					g = cierectolinear(g,1);
					g = (g - botlevel[G]) / (toplevel[G] - botlevel[G]);
					if (g<0) g = 0;
					if (g>1) g = 1;
					g = lineartocierec(g,1);
					out[y][x].g = g*maxpix;
				} else {
					out[y][x].g = in[y][x].g;
				}
				if (enable[B]) {
					b = (double)in[y][x].b / maxpix;
					b = cierectolinear(b,2);
					b = (b - botlevel[B]) / (toplevel[B] - botlevel[B]);
					if (b<0) b = 0;
					if (b>1) b = 1;
					b = lineartocierec(b,2);
					out[y][x].b = b*maxpix;
				} else {
					out[y][x].b = in[y][x].b;
				}
			}
		}
	}

	pnm_freearray(in,ysize);
	if (!outputhist) {
		pnm_writepnm(stdout,out,xsize,ysize,maxpix,format,0/*forceplain*/);
		pnm_freearray(out,ysize);
	} else {
		for (c=0; c<3; c++) {
			if (!enable[c]) continue;
			printf("%s, top=darkest bottom=brightest\n",colname[c]);
			for (i=0; i<histlevels; i++) {
				for (x=0; x<hist[c][i]; x+=(totpix / 130)) {
					printf("*");
				}
				printf("\n");
			}
		}
	}
	exit(EXIT_SUCCESS);
	return 0;
}

